#############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""interfaces for zc.catalog

$Id: interfaces.py 2918 2005-07-19 22:12:38Z jim $
"""

from zope import interface, schema
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
import zope.index.interfaces
import zope.catalog.interfaces
from zc.catalog.i18n import _
import BTrees.Interfaces


class IExtent(interface.Interface):
    """An extent represents the full set of objects indexed by a catalog.
    It is useful for a variety of index operations and catalog queries.
    """

    __parent__ = interface.Attribute(
        """The catalog for which this is an extent; must be None before it is
        set to a catalog""")

    def add(uid, obj):
        """add uid to extent; raise ValueError if it is not addable.

        If uid is already a member of the extent, calling add is a no-op,
        except that if the uid and obj are no longer addable to the extent then
        ValueError is still raised (but without removing the uid)"""

    def remove(uid):
        """Remove uid from set.  Raise KeyError if not a member"""

    def discard(uid):
        """Remove uid from set.  Ignore if not a member"""

    def clear():
        """Remove all uids from set."""

    def __len__():
        """the number of items in the extent."""

    def __iter__():
        """return iterator of uids in set"""

    def __or__(other):
        "Given BTrees.IFBTree data structure, return weighted union"

    def __ror__(other):
        "Given BTrees.IFBTree data structure, return weighted union"

    def union(other, self_weight, other_weight):
        "Given BTrees.IFBTree data structure, return weighted union"

    def __and__(other):
        "Given BTrees.IFBTree data structure, return weighted intersection"

    def __rand__(other):
        "Given BTrees.IFBTree data structure, return weighted intersection"

    def intersection(other, self_weight, other_weight):
        "Given BTrees.IFBTree data structure, return weighted intersection"

    def __sub__(other):
        "extent - set: given BTrees.IFBTree data structure, return difference"

    def difference(other):
        "extent - set: given BTrees.IFBTree data structure, return difference"

    def __rsub__(other):
        "set - extent: given BTrees.IFBTree data structure, return difference"

    def rdifference(other):
        "set - extent: given BTrees.IFBTree data structure, return difference"

    def __nonzero__():
        "return boolean indicating if any uids are in set"

    def __contains__(uid):
        "return boolean indicating if uid is in set"


class IFilterExtent(IExtent):

    filter = interface.Attribute(
        """A (persistent) callable that is passed the extent, a docid, and the
        associated obj and should return a boolean True (is member of extent)
        or False (is not member of extent).""")

    def addable(uid, obj):
        """returns True or False, indicating whether the obj may be added to
        the extent"""


class ISelfPopulatingExtent(IExtent):
    """An extent that knows how to create it's own initial population."""

    populated = schema.Bool(
        title=_("Populated"),
        description=_(
            "Flag indicating whether self-population has been performed."),
        readonly=True,
        )

    def populate():
        """Populate the extent based on the current content of the database.

        After a successful call, `populated` will be True.  Unsuccessful calls
        must raise exceptions.

        If `populated` is true when called, this is a no-op.  After the
        initial population, updates should be maintained via other mechanisms.

        """


class IExtentCatalog(interface.Interface):
    """A catalog of only items within an extent.

    Interface intended to be used with zope.catalog.interfaces.ICatalog"""

    extent = interface.Attribute(
        """An IExtent of the objects cataloged""")


class IIndexValues(interface.Interface):
    """An index that allows introspection of the indexed values"""

    def minValue(min=None):
        """return the minimum value in the index.

        if min is provided, return the minimum value equal to or greater than
        min.

        Raises ValueError if no min.
        """

    def maxValue(max=None):
        """return the maximum value in the index.

        If max is provided, return the maximum value equal to or less than max.

        Raises ValueError if no max.
        """

    def values(min=None, max=None, excludemin=False, excludemax=False,
               doc_id=None):
        """return an iterables of the values in the index.

        if doc_id is provided, returns the values only for that document id.

        If a min is specified, then output is constrained to values greater
        than or equal to the given min, and, if excludemin is specified and
        true, is further constrained to values strictly greater than min.  A
        min value of None is ignored.  If min is None or not specified, and
        excludemin is true, the smallest value is excluded.

        If a max is specified, then output is constrained to values less than
        or equal to the given max, and, if excludemax is specified and
        true, is further constrained to values strictly less than max.  A max
        value of None is ignored.  If max is None or not specified, and
        excludemax is true, the largest value is excluded.
        """

    def containsValue(value):
        """whether the value is used in any of the documents in the index"""

    def ids():
        """return a BTrees.IFBTree data structure of the document ids in the
        index--the ones that have values to be indexed.  All document ids
        should produce at least one value given a call of
        IIndexValues.values(doc_id=id).
        """


class ISetIndex(interface.Interface):

    def apply(query):
        """Return None or an IFBTree Set of the doc ids that match the query.

        query is a dict with one of the following keys: any_of, any,
        all_of, between, and none.

        Any one of the keys may be used; using more than one is not allowed.

        The any_of key should have a value of an iterable of values: the
        result will be the docids whose values contain any of the given values.

        The all_of key should have a value of an iterable of values: the
        result will be the docids whose values contain all of the given values.

        The between key should have a value of an iterable of one to four
        members. The first is the minimum value, or None; the second is the
        maximum value, or None; the third is boolean, defaulting to False,
        declaring if the min should be excluded; and the last is also boolean,
        defaulting to False, declaring if the max should be excluded.

        The any key should take None or an extent.  If the key is None, the
        results will be all docids with any value.  If the key is an extent,
        the results will be the intersection of the extent and all docids with
        any value.

        The none key should take an extent.  It returns the docids in
        the extent that do not have any values in the index.
        """


class IValueIndex(interface.Interface):

    def apply(query):
        """Return None or an IFBTree Set of the doc ids that match the query.

        query is a dict with one of the following keys: any_of, any,
        between, and none.

        Any one of the keys may be used; using more than one is not allowed.

        The any_of key should have a value of an iterable of values: the
        result will be the docids whose values contain any of the given values.

        The between key should have a value of an iterable of one to four
        members. The first is the minimum value, or None; the second is the
        maximum value, or None; the third is boolean, defaulting to False,
        declaring if the min should be excluded; and the last is also boolean,
        defaulting to False, declaring if the max should be excluded.

        The any key should take None or an extent.  If the key is None, the
        results will be all docids with any value.  If the key is an extent,
        the results will be the intersection of the extent and all docids with
        any value.

        The none key should take an extent.  It returns the docids in
        the extent that do not have any values in the index.
        """


class ICatalogValueIndex(zope.catalog.interfaces.IAttributeIndex,
                         zope.catalog.interfaces.ICatalogIndex):
    """Interface-based catalog value index
    """


class ICatalogSetIndex(zope.catalog.interfaces.IAttributeIndex,
                       zope.catalog.interfaces.ICatalogIndex):
    """Interface-based catalog set index
    """


class INormalizationWrapper(zope.index.interfaces.IInjection,
                            zope.index.interfaces.IIndexSearch,
                            zope.index.interfaces.IStatistics,
                            IIndexValues):
    """A wrapper for an index that uses a normalizer to normalize injection
    and querying."""

    index = interface.Attribute(
        """an index implementing IInjection, IIndexSearch, IStatistics, and
        IIndexValues""")

    normalizer = interface.Attribute("a normalizer, implementing INormalizer")

    collection_index = interface.Attribute(
        """boolean: whether indexed values should be treated as collections
        (each composite value normalized) or not (original value is
        normalized)""")


class INormalizer(interface.Interface):

    def value(value):
        """normalize or check constraints for an input value; raise an error
        or return the value to be indexed."""

    def any(value, index):
        """normalize a query value for a "any_of" search; return a sequence of
        values."""

    def all(value, index):
        """Normalize a query value for an "all_of" search; return the value
        for query"""

    def minimum(value, index, exclude=False):
        """normalize a query value for minimum of a range; return the value for
        query"""

    def maximum(value, index, exclude=False):
        """normalize a query value for maximum of a range; return the value for
        query"""

resolution_vocabulary = SimpleVocabulary([SimpleTerm(i, t, t) for i, t in enumerate(
    (_('day'), _('hour'), _('minute'), _('second'), _('microsecond')))])
    #  0         1          2            3            4


class IDateTimeNormalizer(INormalizer):
    resolution = schema.Choice(
        vocabulary=resolution_vocabulary,
        title=_('Resolution'),
        default=2,
        required=True)


class ICallableWrapper(zope.index.interfaces.IInjection,
                       zope.index.interfaces.IIndexSearch,
                       zope.index.interfaces.IStatistics,
                       IIndexValues):
    """A wrapper for an index that uses a callable to convert injection."""

    index = interface.Attribute(
        """An index implementing IInjection, IIndexSearch, IStatistics, and
        IIndexValues""")

    converter = interface.Attribute("A callable converter")



#############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""indexes, as might be found in zope.index

$Id: index.py 2918 2005-07-19 22:12:38Z jim $
"""
import sys
import datetime
import pytz.reference
import BTrees
import persistent
from BTrees import Length
from BTrees.Interfaces import IMerge

from zope import component, interface
import zope.component.interfaces
import zope.interface.common.idatetime
import zope.index.interfaces
from zope.index.field.sorting import SortingIndexMixin
import zope.security.management
from zope.publisher.interfaces import IRequest
import zc.catalog.interfaces
from zc.catalog.i18n import _


class FamilyProperty(object):

    __name__ = "family"

    def __get__(self, instance, type=None):
        if instance is None:
            return self
        d = instance.__dict__
        if "family" in d:
            return d["family"]
        if "btreemodule" in d:
            iftype = d["btreemodule"].split(".")[-1][:2]
            if iftype == "IF":
                d["family"] = BTrees.family32
            elif iftype == "LF":
                d["family"] = BTrees.family64
            else:
                raise ValueError("can't determine btree family based on"
                                 " btreemodule of %r" % (iftype,))
        else:
            d["family"] = BTrees.family32
        self._clear_old_cruft(instance)
        return d["family"]

    def __set__(self, instance, value):
        instance.__dict__["family"] = value
        self._clear_old_cruft(instance)

    def _clear_old_cruft(self, instance):
        d = instance.__dict__
        if "btreemodule" in d:
            del d["btreemodule"]
        if "IOBTree" in d:
            del d["IOBTree"]
        if "BTreeAPI" in d:
            del d["BTreeAPI"]


class AbstractIndex(persistent.Persistent):

    interface.implements(zope.index.interfaces.IInjection,
                         zope.index.interfaces.IIndexSearch,
                         zope.index.interfaces.IStatistics,
                         zc.catalog.interfaces.IIndexValues,
                         )

    family = FamilyProperty()

    def __init__(self, family=None):
        if family is not None:
            self.family = family
        self.clear()

    # These three are deprecated (they were never interface), but can
    # all be computed from the family attribute:

    @property
    def btreemodule(self):
        return self.family.IF.__name__

    @property
    def BTreeAPI(self):
        return self.family.IF

    @property
    def IOBTree(self):
        return self.family.IO.BTree

    def clear(self):
        self.values_to_documents = self.family.OO.BTree()
        self.documents_to_values = self.family.IO.BTree()
        self.documentCount = Length.Length(0)
        self.wordCount = Length.Length(0)

    def minValue(self, min=None):
        if min is None:
            return self.values_to_documents.minKey()
        else:
            return self.values_to_documents.minKey(min)

    def maxValue(self, max=None):
        if max is None:
            return self.values_to_documents.maxKey()
        else:
            return self.values_to_documents.maxKey(max)

    def values(self, min=None, max=None, excludemin=False, excludemax=False,
               doc_id=None):
        if doc_id is None:
            return iter(self.values_to_documents.keys(
                min, max, excludemin, excludemax))
        else:
            values = self.documents_to_values.get(doc_id)
            if values is None:
                return ()
            else:
                return iter(values.keys(min, max, excludemin, excludemax))

    def containsValue(self, value):
        return bool(self.values_to_documents.has_key(value))

    def ids(self):
        return self.documents_to_values.keys()

def parseQuery(query):
    if isinstance(query, dict):
        if len(query) > 1:
            raise ValueError(
                'may only pass one of key, value pair')
        elif not query:
            return None, None
        query_type, query = query.items()[0]
        query_type = query_type.lower()
    else:
        raise ValueError('may only pass a dict to apply')
    return query_type, query

class ValueIndex(SortingIndexMixin, AbstractIndex):

    interface.implements(zc.catalog.interfaces.IValueIndex)

    # attributes used by sorting mixin
    _sorting_num_docs_attr = 'documentCount'        # Length object
    _sorting_fwd_index_attr = 'values_to_documents' # forward BTree index
    _sorting_rev_index_attr = 'documents_to_values' # reverse BTree index

    def _add_value(self, doc_id, added):
        values_to_documents = self.values_to_documents
        docs = values_to_documents.get(added)
        if docs is None:
            values_to_documents[added] = self.family.IF.TreeSet((doc_id,))
            self.wordCount.change(1)
        else:
            docs.insert(doc_id)

    def index_doc(self, doc_id, value):
        if value is None:
            self.unindex_doc(doc_id)
        else:
            values_to_documents = self.values_to_documents
            documents_to_values = self.documents_to_values
            old = documents_to_values.get(doc_id)
            documents_to_values[doc_id] = value
            if old is None:
                self.documentCount.change(1)
            elif old != value:
                docs = values_to_documents.get(old)
                docs.remove(doc_id)
                if not docs:
                    del values_to_documents[old]
                    self.wordCount.change(-1)
            self._add_value(doc_id, value)

    def unindex_doc(self, doc_id):
        documents_to_values = self.documents_to_values
        value = documents_to_values.get(doc_id)
        if value is not None:
            values_to_documents = self.values_to_documents
            self.documentCount.change(-1)
            del documents_to_values[doc_id]
            docs = values_to_documents.get(value)
            docs.remove(doc_id)
            if not docs:
                del values_to_documents[value]
                self.wordCount.change(-1)

    def apply(self, query): # any_of, any, between, none,
        values_to_documents = self.values_to_documents
        query_type, query = parseQuery(query)
        if query_type is None:
            res = None
        elif query_type == 'any_of':
            res = self.family.IF.multiunion(
                [s for s in (values_to_documents.get(v) for v in query)
                 if s is not None])
        elif query_type == 'any':
            if query is None:
                res = self.family.IF.Set(self.ids())
            else:
                assert zc.catalog.interfaces.IExtent.providedBy(query)
                res = query & self.family.IF.Set(self.ids())
        elif query_type == 'between':
            res = self.family.IF.multiunion(
                [s for s in (values_to_documents.get(v) for v in
                             values_to_documents.keys(*query))
                 if s is not None])
        elif query_type == 'none':
            assert zc.catalog.interfaces.IExtent.providedBy(query)
            res = query - self.family.IF.Set(self.ids())
        else:
            raise ValueError(
                "unknown query type", query_type)
        return res

    def values(self, min=None, max=None, excludemin=False, excludemax=False,
               doc_id=None):
        if doc_id is None:
            return iter(self.values_to_documents.keys(
                min, max, excludemin, excludemax))
        else:
            value = self.documents_to_values.get(doc_id)
            if (value is None or
                min is not None and (
                    value < min or excludemin and value == min) or
                max is not None and (
                    value > max or excludemax and value == max)):
                return ()
            else:
                return (value,)

class SetIndex(AbstractIndex):

    interface.implements(zc.catalog.interfaces.ISetIndex)

    def _add_values(self, doc_id, added):
        values_to_documents = self.values_to_documents
        for v in added:
            docs = values_to_documents.get(v)
            if docs is None:
                values_to_documents[v] = self.family.IF.TreeSet((doc_id,))
                self.wordCount.change(1)
            else:
                docs.insert(doc_id)

    def index_doc(self, doc_id, value):
        new = self.family.OO.TreeSet(v for v in value if v is not None)
        if not new:
            self.unindex_doc(doc_id)
        else:
            values_to_documents = self.values_to_documents
            documents_to_values = self.documents_to_values
            old = documents_to_values.get(doc_id)
            if old is None:
                documents_to_values[doc_id] = new
                self.documentCount.change(1)
                self._add_values(doc_id, new)
            else:
                removed = self.family.OO.difference(old, new)
                added = self.family.OO.difference(new, old)
                for v in removed:
                    old.remove(v)
                    docs = values_to_documents.get(v)
                    docs.remove(doc_id)
                    if not docs:
                        del values_to_documents[v]
                        self.wordCount.change(-1)
                old.update(added)
                self._add_values(doc_id, added)

    def unindex_doc(self, doc_id):
        documents_to_values = self.documents_to_values
        values = documents_to_values.get(doc_id)
        if values is not None:
            values_to_documents = self.values_to_documents
            self.documentCount.change(-1)
            del documents_to_values[doc_id]
            for v in values:
                docs = values_to_documents.get(v)
                docs.remove(doc_id)
                if not docs:
                    del values_to_documents[v]
                    self.wordCount.change(-1)

    def apply(self, query): # any_of, any, between, none, all_of
        values_to_documents = self.values_to_documents
        query_type, query = parseQuery(query)
        if query_type is None:
            res = None
        elif query_type == 'any_of':
            res = self.family.IF.Bucket()
            for v in query:
                _, res = self.family.IF.weightedUnion(
                    res, values_to_documents.get(v))
        elif query_type == 'any':
            if query is None:
                res = self.family.IF.Set(self.ids())
            else:
                assert zc.catalog.interfaces.IExtent.providedBy(query)
                res = query & self.family.IF.Set(self.ids())
        elif query_type == 'all_of':
            res = None
            values = iter(query)
            empty = self.family.IF.TreeSet()
            try:
                res = values_to_documents.get(values.next(), empty)
            except StopIteration:
                res = empty
            while res:
                try:
                    v = values.next()
                except StopIteration:
                    break
                res = self.family.IF.intersection(
                    res, values_to_documents.get(v, empty))
        elif query_type == 'between':
            res = self.family.IF.Bucket()
            for v in values_to_documents.keys(*query):
                _, res = self.family.IF.weightedUnion(
                    res, values_to_documents.get(v))
        elif query_type == 'none':
            assert zc.catalog.interfaces.IExtent.providedBy(query)
            res = query - self.family.IF.Set(self.ids())
        else:
            raise ValueError(
                "unknown query type", query_type)
        return res

class NormalizationWrapper(persistent.Persistent):

    interface.implements(zc.catalog.interfaces.INormalizationWrapper)

    index = normalizer = None
    collection_index = False

    def documentCount(self):
        return self.index.documentCount()

    def wordCount(self):
        return self.index.wordCount()

    def clear(self):
        """see zope.index.interfaces.IInjection.clear"""
        return self.index.clear()

    def __init__(self, index, normalizer, collection_index=False):
        self.index = index
        if zope.index.interfaces.IIndexSort.providedBy(index):
            zope.interface.alsoProvides(self, zope.index.interfaces.IIndexSort)
        self.normalizer = normalizer
        self.collection_index = collection_index

    def index_doc(self, doc_id, value):
        if self.collection_index:
            self.index.index_doc(
                doc_id, (self.normalizer.value(v) for v in value))
        else:
            self.index.index_doc(doc_id, self.normalizer.value(value))

    def unindex_doc(self, doc_id):
        self.index.unindex_doc(doc_id)

    def apply(self, query):
        query_type, query = parseQuery(query)
        if query_type == 'any_of':
            res = set()
            for v in query:
                res.update(self.normalizer.any(v, self.index))
        elif query_type == 'all_of':
            res = [self.normalizer.all(v, self.index) for v in query]
        elif query_type == 'between':
            query = tuple(query) # collect iterators
            len_query = len(query)
            max_exclude = len_query >= 4 and bool(query[3])
            min_exclude = len_query >= 3 and bool(query[2])
            max = len_query >= 2 and query[1] and self.normalizer.maximum(
                query[1], self.index, max_exclude) or None
            min = len_query >= 1 and query[0] and self.normalizer.minimum(
                query[0], self.index, min_exclude) or None
            res = (min, max, min_exclude, max_exclude)
        else:
            res = query
        return self.index.apply({query_type: res})

    def minValue(self, min=None):
        if min is not None:
            min = self.normalizer.minimum(min, self.index)
        return self.index.minValue(min)

    def maxValue(self, max=None):
        if max is not None:
            max = self.normalizer.maximum(max, self.index)
        return self.index.maxValue(max)

    def values(self, min=None, max=None, excludemin=False, excludemax=False,
               doc_id=None):
        if min is not None:
            min = self.normalizer.minimum(min, self.index)
        if max is not None:
            max = self.normalizer.maximum(max, self.index)
        return self.index.values(min, max, excludemin, excludemax,
                doc_id=doc_id)

    def containsValue(self, value):
        return self.index.containsValue(value)

    def ids(self):
        return self.index.ids()

    @property
    def sort(self):
        # delegate upstream or raise AttributeError
        return self.index.sort

class CallableWrapper(persistent.Persistent):

    interface.implements(zc.catalog.interfaces.ICallableWrapper)

    converter = None
    index = None

    def __init__(self, index, converter):
        self.index = index
        self.converter = converter

    def index_doc(self, docid, value):
        "See zope.index.interfaces.IInjection"
        self.index.index_doc(docid, self.converter(value))

    def __getattr__(self, name):
        return getattr(self.index, name)


def set_resolution(value, resolution):
    resolution += 2
    if resolution < 6:
        args = []
        args.extend(value.timetuple()[:resolution+1])
        args.extend([0]*(6-resolution))
        args.append(value.tzinfo)
        value = datetime.datetime(*args)
    return value

def get_request():
    i = zope.security.management.queryInteraction()
    if i is not None:
        for p in i.participations:
            if IRequest.providedBy(p):
                return p
    return None

def get_tz(default=pytz.reference.Local):
    request = get_request()
    if request is None:
        return default
    return zope.interface.common.idatetime.ITZInfo(request, default)

def add_tz(value):
    if type(value) is datetime.datetime:
        if value.tzinfo is None:
            value = value.replace(tzinfo=get_tz())
        return value
    else:
        raise ValueError(value)

def day_end(value):
    return (
        datetime.datetime.combine(
            value, datetime.time(tzinfo=get_tz())) +
        datetime.timedelta(days=1) - # separate for daylight savings
        datetime.timedelta(microseconds=1))

def day_begin(value):
    return datetime.datetime.combine(
        value, datetime.time(tzinfo=get_tz()))


class DateTimeNormalizer(persistent.Persistent):

    interface.implements(zc.catalog.interfaces.IDateTimeNormalizer)
    def __init__(self, resolution=2):
        self.resolution = resolution
        # 0, 1, 2, 3, 4
        # day, hour, minute, second, microsecond

    def value(self, value):
        if not isinstance(value, datetime.datetime) or value.tzinfo is None:
            raise ValueError(
                _('This index only indexes timezone-aware datetimes.'))
        return set_resolution(value, self.resolution)

    def any(self, value, index):
        if type(value) is datetime.date:
            start = datetime.datetime.combine(
                value, datetime.time(tzinfo=get_tz()))
            stop = start + datetime.timedelta(days=1)
            return index.values(start, stop, False, True)
        return (add_tz(value),)

    def all(self, value, index):
        return add_tz(value)

    def minimum(self, value, index, exclude=False):
        if type(value) is datetime.date:
            if exclude:
                return day_end(value)
            else:
                return day_begin(value)
        return add_tz(value)

    def maximum(self, value, index, exclude=False):
        if type(value) is datetime.date:
            if exclude:
                return day_begin(value)
            else:
                return day_end(value)
        return add_tz(value)

@interface.implementer(
    zope.interface.implementedBy(NormalizationWrapper),
    zope.index.interfaces.IIndexSort,
    zc.catalog.interfaces.IValueIndex)
def DateTimeValueIndex(resolution=2): # 2 == minute; note that hour is good
    # for timezone-aware per-day searches
    ix = NormalizationWrapper(ValueIndex(), DateTimeNormalizer(resolution))
    interface.alsoProvides(ix, zc.catalog.interfaces.IValueIndex)
    return ix

@interface.implementer(
    zope.interface.implementedBy(NormalizationWrapper),
    zc.catalog.interfaces.ISetIndex)
def DateTimeSetIndex(resolution=2): # 2 == minute; note that hour is good
    # for timezone-aware per-day searches
    ix = NormalizationWrapper(SetIndex(), DateTimeNormalizer(resolution), True)
    interface.alsoProvides(ix, zc.catalog.interfaces.ISetIndex)    
    return ix



#############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Indexes appropriate for zope.catalog

$Id: catalogindex.py 2918 2005-07-19 22:12:38Z jim $
"""
import zope.interface

import zope.catalog.attribute
import zope.container.contained
import zope.index.interfaces

import zc.catalog.index
import zc.catalog.interfaces

class ValueIndex(zope.catalog.attribute.AttributeIndex,
                 zc.catalog.index.ValueIndex,
                 zope.container.contained.Contained):

    zope.interface.implements(zc.catalog.interfaces.ICatalogValueIndex)

class SetIndex(zope.catalog.attribute.AttributeIndex,
               zc.catalog.index.SetIndex,
               zope.container.contained.Contained):

    zope.interface.implements(zc.catalog.interfaces.ICatalogSetIndex)

class NormalizationWrapper(
    zope.catalog.attribute.AttributeIndex,
    zc.catalog.index.NormalizationWrapper,
    zope.container.contained.Contained):

    pass


class CallableWrapper(zc.catalog.index.CallableWrapper,
                      zope.container.contained.Contained):
    zope.interface.implements(zc.catalog.interfaces.ICallableWrapper)


@zope.interface.implementer(
    zope.interface.implementedBy(NormalizationWrapper),
    zc.catalog.interfaces.IValueIndex,
    zope.index.interfaces.IIndexSort)
def DateTimeValueIndex(
    field_name=None, interface=None, field_callable=False,
    resolution=2): # hour; good for per-day searches
    ix = NormalizationWrapper(
        field_name, interface, field_callable, zc.catalog.index.ValueIndex(),
        zc.catalog.index.DateTimeNormalizer(resolution), False)
    zope.interface.alsoProvides(ix, zc.catalog.interfaces.IValueIndex)
    return ix

@zope.interface.implementer(
    zope.interface.implementedBy(NormalizationWrapper),
    zc.catalog.interfaces.ISetIndex)
def DateTimeSetIndex(
    field_name=None, interface=None, field_callable=False, 
    resolution=2): # hour; good for per-day searches
    ix = NormalizationWrapper(
        field_name, interface, field_callable, zc.catalog.index.SetIndex(),
        zc.catalog.index.DateTimeNormalizer(resolution), True)
    zope.interface.alsoProvides(ix, zc.catalog.interfaces.ISetIndex)
    return ix



#############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""extent catalog

$Id: extentcatalog.py 3296 2005-09-09 19:29:20Z benji $
"""

import sys
import BTrees
import persistent
from zope import interface, component
from zope.catalog import catalog
from zope.intid.interfaces import IIntIds

import zope.component
from zope.component.interfaces import IFactory
from BTrees.Interfaces import IMerge

import zc.catalog
from zc.catalog import interfaces


class Extent(persistent.Persistent):
    interface.implements(interfaces.IExtent)
    __parent__ = None
    family = BTrees.family32

    def __init__(self, family=None):
        if family is not None:
            self.family = family
        self.set = self.family.IF.TreeSet()

    # Deprecated.
    @property
    def BTreeAPI(self):
        return sys.modules[self.set.__class__.__module__]

    def __len__(self):
        return len(self.set)

    def add(self, uid, obj):
        self.set.insert(uid)

    def clear(self):
        self.set.clear()

    def __or__(self, other):
        "extent | set"
        return self.union(other)

    __ror__ = __or__

    def union(self, other, self_weight=1, other_weight=1):
        return self.family.IF.weightedUnion(
            self.set, other, self_weight, other_weight)[1]

    def __and__(self, other):
        "extent & set"
        return self.intersection(other)

    __rand__ = __and__

    def intersection(self, other, self_weight=1, other_weight=1):
        return self.family.IF.weightedIntersection(
            self.set, other, self_weight, other_weight)[1]

    def __sub__(self, other):
        "extent - set"
        return self.difference(other)

    def difference(self, other):
        return self.family.IF.difference(self.set, other)

    def __rsub__(self, other):
        "set - extent"
        return self.rdifference(other)

    def rdifference(self, other):
        return self.family.IF.difference(other, self.set)

    def __iter__(self):
        return iter(self.set)

    def __nonzero__(self):
        return bool(self.set)

    def __contains__(self, uid):
        return self.set.has_key(uid)

    def remove(self, uid):
        self.set.remove(uid)

    def discard(self, uid):
        try:
            self.set.remove(uid)
        except KeyError:
            pass


class FilterExtent(Extent):
    interface.implements(interfaces.IFilterExtent)

    def __init__(self, filter, family=None):
        super(FilterExtent, self).__init__(family=family)
        self.filter = filter

    def add(self, uid, obj):
        if not self.addable(uid, obj):
            raise ValueError
        else:
            self.set.insert(uid)

    def addable(self, uid, obj):
        return self.filter(self, uid, obj)


class NonPopulatingExtent(Extent):
    """Base class for populating extent.

    This simple, no-op implementation comes in handy surprisingly often
    for catalogs that handle a very contained domain within an application.
    """

    interface.implements(interfaces.ISelfPopulatingExtent)

    populated = False

    def populate(self):
        self.populated = True


class Catalog(catalog.Catalog):
    interface.implements(interfaces.IExtentCatalog)

    UIDSource = None

    def __init__(self, extent, UIDSource=None):
        """Construct a catalog based on an extent.

        Note that the `family` keyword parameter of the base class
        constructor is not supported here; the family of the extent is
        used.

        """

        self.UIDSource = UIDSource

        if extent.__parent__ is not None:
            raise ValueError("extent's __parent__ must be None")
        super(Catalog, self).__init__(family=extent.family)
        self.extent = extent
        extent.__parent__ = self # inform extent of catalog

    def _getUIDSource(self):
        res = self.UIDSource
        if res is None:
            res = zope.component.getUtility(IIntIds)
        return res

    def clear(self):
        self.extent.clear()
        super(Catalog, self).clear()

    def index_doc(self, docid, texts):
        """Register the data in indexes of this catalog.
        """
        try:
            self.extent.add(docid, texts)
        except ValueError:
            self.unindex_doc(docid)
        else:
            super(Catalog, self).index_doc(docid, texts)

    def unindex_doc(self, docid):
        if docid in self.extent:
            super(Catalog, self).unindex_doc(docid)
            self.extent.remove(docid)

    def updateIndex(self, index):
        if index.__parent__ is not self:
            # not an index in us.  Let the superclass handle it.
            super(Catalog, self).updateIndex(index)
        else:
            uidutil = self._getUIDSource()

            if interfaces.ISelfPopulatingExtent.providedBy(self.extent):
                if not self.extent.populated:
                    self.extent.populate()
                    assert self.extent.populated

                for uid in self.extent:
                    obj = uidutil.getObject(uid)
                    index.index_doc(uid, obj)

            else:
                for uid in uidutil:
                    obj = uidutil.getObject(uid)
                    try:
                        self.extent.add(uid, obj)
                    except ValueError:
                        self.unindex_doc(uid)
                    else:
                        index.index_doc(uid, obj)

    def updateIndexes(self):
        uidutil = self._getUIDSource()

        if interfaces.ISelfPopulatingExtent.providedBy(self.extent):
            if not self.extent.populated:
                self.extent.populate()
                assert self.extent.populated

            for uid in self.extent:
                self.index_doc(uid, uidutil.getObject(uid))

        else:
            for uid in uidutil:
                self.index_doc(uid, uidutil.getObject(uid))
